"""Added CoordSet editable fields

Peek Plugin Database Migration Script

Revision ID: 81c1d2809046
Revises: bf8f559c4df9
Create Date: 2019-05-13 21:08:31.486845

"""

# revision identifiers, used by Alembic.
revision = '81c1d2809046'
down_revision = 'bf8f559c4df9'
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.alter_column('ModelCoordSet', 'enabled', type_=sa.Boolean(), server_default='true',
                    schema='pl_diagram')

    op.execute('UPDATE "pl_diagram"."ModelCoordSet" SET "enabled" = true ')

    op.add_column('ModelCoordSet',
                  sa.Column('editEnabled', sa.Boolean(), server_default='false',
                            nullable=True), schema='pl_diagram')

    op.execute('UPDATE "pl_diagram"."ModelCoordSet" SET "editEnabled" = false ')

    op.alter_column('ModelCoordSet', 'editEnabled', type_=sa.Boolean(), nullable=False,
                    schema='pl_diagram')

    op.add_column('ModelCoordSet',
                  sa.Column('branchesEnabled', sa.Boolean(), server_default='false',
                            nullable=True), schema='pl_diagram')

    op.execute('UPDATE "pl_diagram"."ModelCoordSet" SET "branchesEnabled" = false')

    op.alter_column('ModelCoordSet', 'branchesEnabled', type_=sa.Boolean(),
                    nullable=False,
                    schema='pl_diagram')

    op.add_column('ModelCoordSet',
                  sa.Column('editDefaultColorId', sa.Integer(), nullable=True),
                  schema='pl_diagram')

    op.add_column('ModelCoordSet',
                  sa.Column('editDefaultLayerId', sa.Integer(), nullable=True),
                  schema='pl_diagram')

    op.add_column('ModelCoordSet',
                  sa.Column('editDefaultLevelId', sa.Integer(), nullable=True),
                  schema='pl_diagram')

    op.add_column('ModelCoordSet',
                  sa.Column('editDefaultLineStyleId', sa.Integer(), nullable=True),
                  schema='pl_diagram')

    op.add_column('ModelCoordSet',
                  sa.Column('editDefaultTextStyleId', sa.Integer(), nullable=True),
                  schema='pl_diagram')

    op.create_index('idxCoordModel_editDefaultColorId', 'ModelCoordSet',
                    ['editDefaultColorId'], unique=False, schema='pl_diagram')

    op.create_index('idxCoordModel_editDefaultLayerId', 'ModelCoordSet',
                    ['editDefaultLayerId'], unique=False, schema='pl_diagram')

    op.create_index('idxCoordModel_editDefaultLevelId', 'ModelCoordSet',
                    ['editDefaultLevelId'], unique=False, schema='pl_diagram')

    op.create_index('idxCoordModel_editDefaultLineStyleId', 'ModelCoordSet',
                    ['editDefaultLineStyleId'], unique=False, schema='pl_diagram')

    op.create_index('idxCoordModel_editDefaultTextStyleId', 'ModelCoordSet',
                    ['editDefaultTextStyleId'], unique=False, schema='pl_diagram')

    op.create_foreign_key('ModelCoordSet_editDefaultLineStyleId_fkey', 'ModelCoordSet',
                          'DispLineStyle',
                          ['editDefaultLineStyleId'], ['id'], source_schema='pl_diagram',
                          referent_schema='pl_diagram')

    op.create_foreign_key('ModelCoordSet_editDefaultColorId_fkey', 'ModelCoordSet',
                          'DispColor', ['editDefaultColorId'],
                          ['id'], source_schema='pl_diagram',
                          referent_schema='pl_diagram')

    op.create_foreign_key('ModelCoordSet_editDefaultLayerId_fkey', 'ModelCoordSet',
                          'DispLayer', ['editDefaultLayerId'],
                          ['id'], source_schema='pl_diagram',
                          referent_schema='pl_diagram')

    op.create_foreign_key('ModelCoordSet_editDefaultLevelId_fkey', 'ModelCoordSet',
                          'DispLevel', ['editDefaultLevelId'],
                          ['id'], source_schema='pl_diagram',
                          referent_schema='pl_diagram')

    op.create_foreign_key('ModelCoordSet_editDefaultTextStyleId_fkey', 'ModelCoordSet',
                          'DispTextStyle',
                          ['editDefaultTextStyleId'], ['id'], source_schema='pl_diagram',
                          referent_schema='pl_diagram')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint('ModelCoordSet_editDefaultLineStyleId_fkey', 'ModelCoordSet',
                       schema='pl_diagram', type_='foreignkey')
    op.drop_constraint('ModelCoordSet_editDefaultColorId_fkey', 'ModelCoordSet',
                       schema='pl_diagram', type_='foreignkey')
    op.drop_constraint('ModelCoordSet_editDefaultLayerId_fkey', 'ModelCoordSet',
                       schema='pl_diagram', type_='foreignkey')
    op.drop_constraint('ModelCoordSet_editDefaultLevelId_fkey', 'ModelCoordSet',
                       schema='pl_diagram', type_='foreignkey')
    op.drop_constraint('ModelCoordSet_editDefaultTextStyleId_fkey', 'ModelCoordSet',
                       schema='pl_diagram', type_='foreignkey')

    op.drop_index('idxCoordModel_editDefaultTextStyleId', table_name='ModelCoordSet',
                  schema='pl_diagram')
    op.drop_index('idxCoordModel_editDefaultLineStyleId', table_name='ModelCoordSet',
                  schema='pl_diagram')
    op.drop_index('idxCoordModel_editDefaultLevelId', table_name='ModelCoordSet',
                  schema='pl_diagram')
    op.drop_index('idxCoordModel_editDefaultLayerId', table_name='ModelCoordSet',
                  schema='pl_diagram')
    op.drop_index('idxCoordModel_editDefaultColorId', table_name='ModelCoordSet',
                  schema='pl_diagram')
    op.drop_column('ModelCoordSet', 'branchesEnabled', schema='pl_diagram')
    op.drop_column('ModelCoordSet', 'editEnabled', schema='pl_diagram')
    op.drop_column('ModelCoordSet', 'editDefaultTextStyleId', schema='pl_diagram')
    op.drop_column('ModelCoordSet', 'editDefaultLineStyleId', schema='pl_diagram')
    op.drop_column('ModelCoordSet', 'editDefaultLevelId', schema='pl_diagram')
    op.drop_column('ModelCoordSet', 'editDefaultLayerId', schema='pl_diagram')
    op.drop_column('ModelCoordSet', 'editDefaultColorId', schema='pl_diagram')
    # ### end Alembic commands ###



from datetime import datetime

from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from vortex.Tuple import addTupleType, TupleField, Tuple


@addTupleType
class GridTuple(Tuple):
    __tupleType__ = diagramTuplePrefix + "GridTuple"

    gridKey: str = TupleField()
    dispJsonStr: str = TupleField()
    lastUpdate: datetime = TupleField()


import logging
from datetime import datetime
from typing import List, Callable

import pytz
from sqlalchemy import asc
from twisted.internet import task, reactor, defer
from twisted.internet.defer import inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger, vortexLogFailure

from peek_plugin_diagram._private.server.client_handlers.ClientLocationIndexUpdateHandler import \
    ClientLocationIndexUpdateHandler
from peek_plugin_diagram._private.server.controller.StatusController import \
    StatusController
from peek_plugin_diagram._private.storage.LocationIndex import \
    LocationIndexCompilerQueue

logger = logging.getLogger(__name__)


class DispKeyCompilerQueueController:
    """ Disp Compiler

    Compile the disp items into the grid data

    1) Query for queue
    2) Process queue
    3) Delete from queue

    """

    ITEMS_PER_TASK = 10
    PERIOD = 5.000

    QUEUE_MAX = 10
    QUEUE_MIN = 0

    TASK_TIMEOUT = 60.0

    def __init__(self, ormSessionCreator,
                 statusController: StatusController,
                 clientLocationUpdateHandler: ClientLocationIndexUpdateHandler,
                 readyLambdaFunc: Callable):
        self._dbSessionCreator = ormSessionCreator
        self._statusController: StatusController = statusController
        self._clientLocationUpdateHandler: ClientLocationIndexUpdateHandler = clientLocationUpdateHandler
        self._readyLambdaFunc = readyLambdaFunc

        self._pollLoopingCall = task.LoopingCall(self._poll)
        self._lastQueueId = -1
        self._queueCount = 0

        self._chunksInProgress = set()

    def start(self):
        self._statusController.setLocationIndexCompilerStatus(True, self._queueCount)
        d = self._pollLoopingCall.start(self.PERIOD, now=False)
        d.addCallbacks(self._timerCallback, self._timerErrback)

    def _timerErrback(self, failure):
        vortexLogFailure(failure, logger)
        self._statusController.setLocationIndexCompilerStatus(False, self._queueCount)
        self._statusController.setLocationIndexCompilerError(str(failure.value))

    def _timerCallback(self, _):
        self._statusController.setLocationIndexCompilerStatus(False, self._queueCount)

    def stop(self):
        if self._pollLoopingCall.running:
            self._pollLoopingCall.stop()

    def shutdown(self):
        self.stop()

    @inlineCallbacks
    def _poll(self):
        if not self._readyLambdaFunc():
            return

        # We queue the grids in bursts, reducing the work we have to do.
        if self._queueCount > self.QUEUE_MIN:
            return

        # Check for queued items
        queueItems = yield self._grabQueueChunk()
        if not queueItems:
            return

        # Send the tasks to the peek worker
        for start in range(0, len(queueItems), self.ITEMS_PER_TASK):

            items = queueItems[start: start + self.ITEMS_PER_TASK]

            # If we're already processing these chunks, then return and try later
            if self._chunksInProgress & set([o.indexBucket for o in items]):
                return

            # Set the watermark
            self._lastQueueId = items[-1].id

            # This should never fail
            d = self._sendToWorker(items)
            d.addErrback(vortexLogFailure, logger)

            self._queueCount += 1
            if self._queueCount >= self.QUEUE_MAX:
                break

        yield self._dedupeQueue()

    @inlineCallbacks
    def _sendToWorker(self, items: List[LocationIndexCompilerQueue]):
        from peek_plugin_diagram._private.worker.tasks.LocationIndexCompilerTask import \
            compileLocationIndex

        startTime = datetime.now(pytz.utc)

        # Add the chunks we're processing to the set
        self._chunksInProgress |= set([o.indexBucket for o in items])

        try:
            d = compileLocationIndex.delay(items)
            d.addTimeout(self.TASK_TIMEOUT, reactor)

            indexBuckets = yield d
            logger.debug("Time Taken = %s" % (datetime.now(pytz.utc) - startTime))

            self._queueCount -= 1

            self._clientLocationUpdateHandler.sendLocationIndexes(indexBuckets)
            self._statusController.addToLocationIndexCompilerTotal(len(items))
            self._statusController.setLocationIndexCompilerStatus(True, self._queueCount)

            # Success, Remove the chunks from the in-progress queue
            self._chunksInProgress -= set([o.indexBucket for o in items])

        except Exception as e:
            if isinstance(e, defer.TimeoutError):
                logger.info("Retrying compile, Task has timed out.")
            else:
                logger.debug("Retrying compile : %s", str(e))

            reactor.callLater(2.0, self._sendToWorker, items)
            return

    @deferToThreadWrapWithLogger(logger)
    def _grabQueueChunk(self):
        session = self._dbSessionCreator()
        try:
            qry = session.query(LocationIndexCompilerQueue) \
                .order_by(asc(LocationIndexCompilerQueue.id)) \
                .filter(LocationIndexCompilerQueue.id > self._lastQueueId) \
                .yield_per(self.QUEUE_MAX) \
                .limit(self.QUEUE_MAX)

            queueItems = qry.all()
            session.expunge_all()

            # Deduplicate the values and return the ones with the lowest ID
            return list({o.indexBucket: o for o in reversed(queueItems)}.values())

        finally:
            session.close()

    @deferToThreadWrapWithLogger(logger)
    def _dedupeQueue(self):
        session = self._dbSessionCreator()
        try:
            session.execute("""
                with sq as (
                    SELECT min(id) as "minId"
                    FROM pl_diagram."LocationIndexCompilerQueue"
                    WHERE id > %s
                    GROUP BY "modelSetId", "indexBucket"
                )
                DELETE
                FROM pl_diagram."LocationIndexCompilerQueue"
                WHERE "id" not in (SELECT "minId" FROM sq)
            """ % self._lastQueueId)
            session.commit()
        finally:
            session.close()



from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from vortex.Tuple import addTupleType, TupleField, Tuple


@addTupleType
class EncodedLocationIndexTuple(Tuple):
    __tupleType__ = diagramTuplePrefix + "EncodedLocationIndexTuple"

    modelSetKey: str = TupleField()
    indexBucket: str = TupleField()

    # The LocationIndexTuple pre-encoded to a Payload
    encodedLocationIndexTuple: bytes = TupleField()
    lastUpdate: str = TupleField()



from typing import List, Any

from vortex.Tuple import addTupleType, TupleField, Tuple

from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix


@addTupleType
class GroupDispsTuple(Tuple):
    """ Group Disps Tuple

    This tuple stores a list of DispGroups that are in the 'ID:dispgroup' grid key
    in that coord set.

    """
    __tupleType__ = diagramTuplePrefix + "GroupDispsTuple"

    coordSetId: int = TupleField()

    # A GridTuple, already encoded and ready for storage in the clients cache
    encodedGridTuple: str = TupleField()



import logging
from typing import List

from sqlalchemy import select
from twisted.internet.defer import Deferred, inlineCallbacks

from peek_plugin_base.storage.StorageUtil import makeOrmValuesSubqueryCondition, \
    makeCoreValuesSubqueryCondition
from peek_plugin_diagram._private.storage.GridKeyIndex import GridKeyIndex, \
    DispIndexerQueue
from peek_plugin_diagram._private.storage.LiveDbDispLink import LiveDbDispLink
from peek_plugin_diagram._private.worker.tasks.LiveDbDisplayValueConverterTask import \
    convertLiveDbDisplayValuesTask
from peek_plugin_livedb.server.LiveDBReadApiABC import LiveDBReadApiABC
from peek_plugin_livedb.server.LiveDBWriteApiABC import LiveDBWriteApiABC
from peek_plugin_livedb.tuples.LiveDbDisplayValueUpdateTuple import \
    LiveDbDisplayValueUpdateTuple
from peek_plugin_livedb.tuples.LiveDbRawValueUpdateTuple import LiveDbRawValueUpdateTuple
from vortex.DeferUtil import deferToThreadWrapWithLogger, \
    vortexInlineCallbacksLogAndConsumeFailure

logger = logging.getLogger(__name__)


class LiveDbWatchController:
    """ Watch Grid Controller

    This controller handles most of the interactions with the LiveDB plugin..

    That is :

    1) Informing the LiveDB that the keys are being watched

    2) Sending updates for these watched grids to the clients (???? is this the right spot?)

    3) Computing display values for the live db

    4) Queueing display updates to be compiled, based on events from the livedb.

    """

    def __init__(self, liveDbWriteApi: LiveDBWriteApiABC,
                 liveDbReadApi: LiveDBReadApiABC,
                 dbSessionCreator):
        self._liveDbWriteApi = liveDbWriteApi
        self._liveDbReadApi = liveDbReadApi
        self._dbSessionCreator = dbSessionCreator

        self._liveDbReadApi.rawValueUpdatesObservable('pofDiagram').subscribe(
            on_next=self.processLiveDbRawValueUpdates
        )
        self._liveDbReadApi.displayValueUpdatesObservable('pofDiagram').subscribe(
            on_next=self.processLiveDbDisplayValueUpdates
        )

    def shutdown(self):
        pass

    @inlineCallbacks
    def updateClientWatchedGrids(self, clientId: str, gridKeys: List[str]) -> Deferred:
        """ Update Client Watched Grids

        Tell the server that these grids are currently being watched by users.

        :param clientId: A unique identifier of the client (Maybe it's vortex uuid)
        :param gridKeys: A list of grid keys that this client is observing.
        :returns: Nothing
        """

        try:
            liveDbKeys = yield self.getLiveDbKeys(gridKeys)
            self._liveDbWriteApi.prioritiseLiveDbValueAcquisition(
                'pofDiagram', liveDbKeys
            )

        except Exception as e:
            logger.exception(e)

    @vortexInlineCallbacksLogAndConsumeFailure(logger)
    def processLiveDbRawValueUpdates(self, updates: List[LiveDbRawValueUpdateTuple]):
        modelSetKey = 'pofDiagram'

        displayValueUpdates = yield convertLiveDbDisplayValuesTask.delay(
            modelSetKey, updates
        )

        self._liveDbWriteApi.updateDisplayValue(modelSetKey, displayValueUpdates)

    @deferToThreadWrapWithLogger(logger)
    def getLiveDbKeys(self, gridKeys) -> List[str]:

        session = self._dbSessionCreator()
        try:
            return [t[0] for t in
                    session.query(LiveDbDispLink.liveDbKey)
                        .join(GridKeyIndex,
                              GridKeyIndex.dispId == LiveDbDispLink.dispId)
                        .filter(makeOrmValuesSubqueryCondition(
                        session, GridKeyIndex.gridKey, gridKeys
                    ))
                        .yield_per(1000)
                        .distinct()]
        finally:
            session.close()

    @deferToThreadWrapWithLogger(logger, consumeError=True)
    def processLiveDbDisplayValueUpdates(self,
                                         updates: List[LiveDbDisplayValueUpdateTuple]):

        logger.debug("TODO TODO - processLiveDbDisplayValueUpdates coordSetId")
        linkTable = LiveDbDispLink.__table__
        queueTable = DispIndexerQueue.__table__

        session = self._dbSessionCreator()
        try:
            updatedKeys = [u.key for u in updates]
            stmt = (select([linkTable.c.dispId])
                .where(makeCoreValuesSubqueryCondition(
                session.bind, linkTable.c.liveDbKey, updatedKeys
            ))
            )

            ins = queueTable.insert().from_select(['dispId'], stmt)
            session.execute(ins)
            session.commit()

        finally:
            session.close()



from sqlalchemy.ext.declarative import declarative_base

from sqlalchemy.schema import MetaData
from txhttputil.util.ModuleUtil import filterModules

from peek_plugin_diagram._private.storage import branch

metadata = MetaData(schema="pl_diagram")
DeclarativeBase = declarative_base(metadata=metadata)


def loadStorageTuples():

    """ Load Storage Tables

    This method should be called from the "load()" method of the agent, server, worker
    and client entry hook classes.

    This will register the ORM classes as tuples, allowing them to be serialised and
    deserialized by the vortex.

    """
    for mod in filterModules(__package__, __file__):
        if mod.startswith("Declarative"):
            continue
        __import__(mod, locals(), globals())

    for mod in filterModules(branch.__package__, branch.__file__):
        __import__(mod, locals(), globals())



from typing import List

from abc import ABCMeta, abstractmethod
from twisted.internet.defer import Deferred


class DiagramViewerApiABC(metaclass=ABCMeta):
    @abstractmethod
    def getCoordSets(self, modelSetKey: str) -> Deferred:
        """ Get Coordinate Set Tuples

        Returns a list of coordinate set tuples for a model set name

        :param modelSetKey:  The name of the model set query for.

        :return: A deferred that fires with a list of tuples
                List[DiagramCoordSetTuple]

        """



"""Added landingCoordSetId

Peek Plugin Database Migration Script

Revision ID: 97990b4ca6a1
Revises: da424430bf97
Create Date: 2019-07-26 13:24:14.505584

"""

# revision identifiers, used by Alembic.
revision = '97990b4ca6a1'
down_revision = 'da424430bf97'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('ModelSet', sa.Column('landingCoordSetId', sa.Integer(), nullable=True), schema='pl_diagram')
    op.create_foreign_key(None, 'ModelSet', 'ModelCoordSet', ['landingCoordSetId'], ['id'], source_schema='pl_diagram', referent_schema='pl_diagram', ondelete='CASCADE')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, 'ModelSet', schema='pl_diagram', type_='foreignkey')
    op.drop_column('ModelSet', 'landingCoordSetId', schema='pl_diagram')
    # ### end Alembic commands ###


import logging
from datetime import datetime
from typing import List

import pytz
from peek_plugin_diagram._private.server.client_handlers.ClientGridUpdateHandler import \
    ClientGridUpdateHandler
from peek_plugin_diagram._private.server.controller.StatusController import \
    StatusController
from peek_plugin_diagram._private.storage.GridKeyIndex import \
    GridKeyCompilerQueue
from sqlalchemy import asc
from twisted.internet import task, reactor, defer
from twisted.internet.defer import inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger, vortexLogFailure

logger = logging.getLogger(__name__)


class GridKeyCompilerQueueController:
    """ Grid Compiler

    Compile the disp items into the grid data

    1) Query for queue
    2) Process queue
    3) Delete from queue

    """

    ITEMS_PER_TASK = 5
    PERIOD = 0.200

    QUEUE_MAX = 100
    QUEUE_MIN = 30

    TASK_TIMEOUT = 60.0

    def __init__(self, ormSessionCreator,
                 statusController: StatusController,
                 clientGridUpdateHandler: ClientGridUpdateHandler):
        self._dbSessionCreator = ormSessionCreator
        self._statusController: StatusController = statusController
        self._clientGridUpdateHandler: ClientGridUpdateHandler = clientGridUpdateHandler

        self._pollLoopingCall = task.LoopingCall(self._poll)
        self._lastQueueId = -1
        self._queueCount = 0

        self._chunksInProgress = set()

    def start(self):
        self._statusController.setGridCompilerStatus(True, self._queueCount)
        d = self._pollLoopingCall.start(self.PERIOD, now=False)
        d.addCallbacks(self._timerCallback, self._timerErrback)

    def isBusy(self) -> bool:
        return self._queueCount != 0

    def _timerErrback(self, failure):
        vortexLogFailure(failure, logger)
        self._statusController.setGridCompilerStatus(False, self._queueCount)
        self._statusController.setGridCompilerError(str(failure.value))

    def _timerCallback(self, _):
        self._statusController.setGridCompilerStatus(False, self._queueCount)

    def stop(self):
        if self._pollLoopingCall.running:
            self._pollLoopingCall.stop()

    def shutdown(self):
        self.stop()

    @inlineCallbacks
    def _poll(self):

        # We queue the grids in bursts, reducing the work we have to do.
        if self._queueCount > self.QUEUE_MIN:
            return

        # Check for queued items
        queueItems = yield self._grabQueueChunk()
        if not queueItems:
            return

        # Send the tasks to the peek worker
        for start in range(0, len(queueItems), self.ITEMS_PER_TASK):

            items = queueItems[start: start + self.ITEMS_PER_TASK]

            # If we're already processing these chunks, then return and try later
            if self._chunksInProgress & set([o.gridKey for o in items]):
                return

            # This should never fail
            d = self._sendToWorker(items)
            d.addErrback(vortexLogFailure, logger)

            # Set the watermark
            self._lastQueueId = items[-1].id

            self._queueCount += 1
            if self._queueCount >= self.QUEUE_MAX:
                break

        yield self._dedupeQueue()

    @inlineCallbacks
    def _sendToWorker(self, items: List[GridKeyCompilerQueue]):
        from peek_plugin_diagram._private.worker.tasks.GridCompilerTask import \
            compileGrids

        startTime = datetime.now(pytz.utc)

        # Add the chunks we're processing to the set
        self._chunksInProgress |= set([o.gridKey for o in items])

        try:
            d = compileGrids.delay(items)
            d.addTimeout(self.TASK_TIMEOUT, reactor)

            gridKeys = yield d
            logger.debug("Time Taken = %s" % (datetime.now(pytz.utc) - startTime))

            self._queueCount -= 1

            self._clientGridUpdateHandler.sendGrids(gridKeys)
            self._statusController.addToGridCompilerTotal(len(items))
            self._statusController.setGridCompilerStatus(True, self._queueCount)

            # Success, Remove the chunks from the in-progress queue
            self._chunksInProgress -= set([o.gridKey for o in items])

        except Exception as e:
            if isinstance(e, defer.TimeoutError):
                logger.info("Retrying compile, Task has timed out.")
            else:
                logger.debug("Retrying compile : %s", str(e))

            reactor.callLater(2.0, self._sendToWorker, items)
            return

    @deferToThreadWrapWithLogger(logger)
    def _grabQueueChunk(self):
        session = self._dbSessionCreator()
        try:
            qry = (session.query(GridKeyCompilerQueue)
                   .order_by(asc(GridKeyCompilerQueue.id))
                   .filter(GridKeyCompilerQueue.id > self._lastQueueId)
                   .yield_per(self.QUEUE_MAX)
                   .limit(self.QUEUE_MAX)
                   )

            queueItems = qry.all()
            session.expunge_all()

            # Deduplicate the values and return the ones with the lowest ID
            return list({o.gridKey: o for o in reversed(queueItems)}.values())

        finally:
            session.close()

    @deferToThreadWrapWithLogger(logger)
    def _dedupeQueue(self):
        session = self._dbSessionCreator()
        try:
            session.execute("""
                with sq as (
                    SELECT min(id) as "minId"
                    FROM pl_diagram."GridKeyCompilerQueue"
                    WHERE id > %s
                    GROUP BY "coordSetId", "gridKey"
                )
                DELETE
                FROM pl_diagram."GridKeyCompilerQueue"
                WHERE "id" not in (SELECT "minId" FROM sq)
            """ % self._lastQueueId)
            session.commit()
        finally:
            session.close()



import logging

from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer, String
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import Index

from peek_plugin_base.storage.TypeDecorators import PeekLargeBinary
from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix
from vortex.Tuple import Tuple, addTupleType
from .DeclarativeBase import DeclarativeBase
from .Display import DispBase
from .ModelSet import ModelSet

logger = logging.getLogger(__name__)


@addTupleType
class LocationIndexCompilerQueue(Tuple, DeclarativeBase):
    __tablename__ = 'LocationIndexCompilerQueue'
    __tupleType__ = diagramTuplePrefix + __tablename__

    id = Column(Integer, primary_key=True, autoincrement=True)

    indexBucket = Column(String(100), primary_key=True)
    modelSetId = Column(Integer,
                        ForeignKey('ModelSet.id', ondelete='CASCADE'),
                        primary_key=True)

    __table_args__ = (
        Index("idx_LICompQueue_modelSetId_indexBucket", modelSetId, indexBucket,
              unique=False),
    )


@addTupleType
class LocationIndex(Tuple, DeclarativeBase):
    __tablename__ = 'LocationIndex'
    __tupleType__ = diagramTuplePrefix + __tablename__

    indexBucket = Column(String(100), primary_key=True)
    dispId = Column(Integer,
                    ForeignKey('DispBase.id', ondelete='CASCADE'),
                    primary_key=True)

    disp = relationship(DispBase)

    modelSetId = Column(Integer, ForeignKey('ModelSet.id'), nullable=False)
    modelSet = relationship(ModelSet)

    __table_args__ = (
        Index("idx_LocationIndex_indexBucket", indexBucket, unique=False),
        Index("idx_LocationIndex_dispId", dispId, unique=False),
        Index("idx_LocationIndex_modelSetId", modelSetId, unique=False),
    )


@addTupleType
class LocationIndexCompiled(Tuple, DeclarativeBase):
    __tablename__ = 'LocationIndexCompiled'
    __tupleType__ = diagramTuplePrefix + __tablename__

    id = Column(Integer, primary_key=True, autoincrement=True)

    indexBucket = Column(String(100), primary_key=True)
    blobData = Column(PeekLargeBinary, nullable=False)
    lastUpdate = Column(String(50), nullable=False)

    modelSetId = Column(Integer,
                        ForeignKey('ModelSet.id', ondelete='CASCADE'),
                        nullable=False)
    modelSet = relationship(ModelSet)

    __table_args__ = (
        Index("idx_LIIndexUpdate_modelSetId", modelSetId, unique=False),
        Index("idx_LIIndexUpdate_indexBucket", indexBucket, unique=True),
    )



"""Added polyline startKey endKey

Peek Plugin Database Migration Script

Revision ID: 558bc7fc4d36
Revises: 24b83c5e7e31
Create Date: 2017-09-23 14:34:32.435996

"""

# revision identifiers, used by Alembic.
revision = '558bc7fc4d36'
down_revision = '24b83c5e7e31'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.add_column('DispPolyline', sa.Column('endKey', sa.String(length=50), nullable=True), schema='pl_diagram')
    op.add_column('DispPolyline', sa.Column('startKey', sa.String(length=50), nullable=True), schema='pl_diagram')
    op.create_index('idx_DispPolyline_endKey', 'DispPolyline', ['endKey'], unique=False, schema='pl_diagram')
    op.create_index('idx_DispPolyline_startKey', 'DispPolyline', ['startKey'], unique=False, schema='pl_diagram')



def downgrade():
    op.drop_index('idx_DispPolyline_startKey', table_name='DispPolyline', schema='pl_diagram')
    op.drop_index('idx_DispPolyline_endKey', table_name='DispPolyline', schema='pl_diagram')
    op.drop_column('DispPolyline', 'startKey', schema='pl_diagram')
    op.drop_column('DispPolyline', 'endKey', schema='pl_diagram')



"""added coordset mx c

Peek Plugin Database Migration Script

Revision ID: 4b943b584307
Revises: d103b9c1eba9
Create Date: 2017-10-01 11:16:34.264774

"""

# revision identifiers, used by Alembic.
revision = '4b943b584307'
down_revision = 'd103b9c1eba9'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.add_column('ModelCoordSet', sa.Column('multiplierX', sa.Float(), server_default='1', nullable=True), schema='pl_diagram')
    op.add_column('ModelCoordSet', sa.Column('multiplierY', sa.Float(), server_default='1', nullable=True), schema='pl_diagram')
    op.add_column('ModelCoordSet', sa.Column('offsetX', sa.Float(), server_default='0', nullable=True), schema='pl_diagram')
    op.add_column('ModelCoordSet', sa.Column('offsetY', sa.Float(), server_default='0', nullable=True), schema='pl_diagram')

    op.execute('''UPDATE "pl_diagram"."ModelCoordSet"
                  SET "multiplierX" = 1,
                        "multiplierY" = 1,
                        "offsetX" = 0,
                        "offsetY" = 0
                       ''')

    op.alter_column('ModelCoordSet', 'multiplierX', type_=sa.Float, nullable=False, schema='pl_diagram')
    op.alter_column('ModelCoordSet', 'multiplierY', type_=sa.Float, nullable=False, schema='pl_diagram')
    op.alter_column('ModelCoordSet', 'offsetX', type_=sa.Float, nullable=False, schema='pl_diagram')
    op.alter_column('ModelCoordSet', 'offsetY', type_=sa.Float, nullable=False, schema='pl_diagram')
    
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('ModelCoordSet', 'offsetY', schema='pl_diagram')
    op.drop_column('ModelCoordSet', 'offsetX', schema='pl_diagram')
    op.drop_column('ModelCoordSet', 'multiplierY', schema='pl_diagram')
    op.drop_column('ModelCoordSet', 'multiplierX', schema='pl_diagram')
    # ### end Alembic commands ###


import logging
from typing import List, Optional

from twisted.internet.defer import Deferred

from peek_plugin_base.PeekVortexUtil import peekClientName
from peek_plugin_base.storage.StorageUtil import makeOrmValuesSubqueryCondition
from peek_plugin_diagram._private.client.controller.GridCacheController import \
    clientGridUpdateFromServerFilt
from peek_plugin_diagram._private.storage.GridKeyIndex import GridKeyIndexCompiled
from peek_plugin_diagram._private.tuples.grid.GridTuple import GridTuple
from peek_plugin_diagram._private.tuples.grid.EncodedGridTuple import EncodedGridTuple
from vortex.DeferUtil import vortexLogFailure, deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.VortexFactory import VortexFactory, NoVortexException

logger = logging.getLogger(__name__)


class ClientGridUpdateHandler:
    """ Client Update Controller

    This controller handles sending updates the the client.

    It uses lower level Vortex API

    It does the following a broadcast to all clients:

    1) Sends grid updates to the clients

    2) Sends Lookup updates to the clients

    """

    def __init__(self, dbSessionCreator):
        self._dbSessionCreator = dbSessionCreator

    def shutdown(self):
        pass

    def sendGrids(self, gridKeys: List[str]) -> None:
        """ Send Grids

        Send grid updates to the client services

        :param gridKeys: A list of grid keys that this client is observing.
                            the grids are sent to the requesting client only.
        :returns: Nothing
        """

        if not gridKeys:
            return

        if peekClientName not in VortexFactory.getRemoteVortexName():
            logger.debug("No clients are online to send the grid to, %s", gridKeys)
            return

        def send(vortexMsg: Optional[bytes]):
            if vortexMsg:
                VortexFactory.sendVortexMsg(
                    vortexMsg, destVortexName=peekClientName
                )

        d: Deferred = self._serialiseGrids(gridKeys)
        d.addCallback(send)
        d.addErrback(self._sendGridsErrback, gridKeys)

    def _sendGridsErrback(self, failure, gridKeys):

        if failure.check(NoVortexException):
            logger.debug(
                "No clients are online to send the grid to, %s", gridKeys)
            return

        vortexLogFailure(failure, logger)

    @deferToThreadWrapWithLogger(logger)
    def _serialiseGrids(self, gridKeys) -> Optional[bytes]:
        session = self._dbSessionCreator()
        try:
            ormGrids = (session.query(GridKeyIndexCompiled)
                        .filter(makeOrmValuesSubqueryCondition(
                            session, GridKeyIndexCompiled.gridKey, gridKeys
                        ))
                        .yield_per(200))

            gridTuples: List[EncodedGridTuple] = []
            for ormGrid in ormGrids:
                gridTuples.append(
                    EncodedGridTuple(gridKey=ormGrid.gridKey,
                                     encodedGridTuple=ormGrid.encodedGridTuple,
                                     lastUpdate=ormGrid.lastUpdate)
                )

            if not gridTuples:
                return None

            return Payload(
                filt=clientGridUpdateFromServerFilt, tuples=gridTuples
                           ).makePayloadEnvelope(compressionLevel=3).toVortexMsg()

        finally:
            session.close()



import logging
from collections import defaultdict
from typing import Union

from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

from peek_plugin_diagram._private.storage.branch.BranchIndex import BranchIndex
from peek_plugin_diagram._private.tuples.branch.BranchKeyToIdMapTuple import \
    BranchKeyToIdMapTuple

logger = logging.getLogger(__name__)


class BranchKeyToIdMapTupleProvider(TuplesProviderABC):
    def __init__(self, ormSessionCreator):
        self._ormSessionCreator = ormSessionCreator

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(self, filt: dict,
                      tupleSelector: TupleSelector) -> Union[Deferred, bytes]:

        session = self._ormSessionCreator()
        try:
            tuples = session.query(BranchIndex.id,
                                   BranchIndex.coordSetId,
                                   BranchIndex.key).all()

            tupleByCoordSetId = defaultdict(BranchKeyToIdMapTuple)
            for t in tuples:
                newTuple = tupleByCoordSetId[t.coordSetId]
                if not newTuple.keyIdMap:
                    newTuple.keyIdMap = {}
                newTuple.keyIdMap[t.key] = t.id
                newTuple.coordSetId = t.coordSetId

            return Payload(filt, tuples=list(tupleByCoordSetId.values())) \
                .makePayloadEnvelope().toVortexMsg()

        finally:
            session.close()



import logging
from typing import List

from twisted.internet.defer import Deferred

from peek_plugin_diagram._private.server.controller.BranchUpdateController import \
    BranchUpdateController
from peek_plugin_diagram._private.server.controller.DispImportController import \
    DispImportController
from peek_plugin_diagram._private.server.controller.LookupImportController import \
    LookupImportController
from peek_plugin_diagram._private.server.controller.StatusController import \
    StatusController
from peek_plugin_diagram.server.DiagramImportApiABC import DiagramImportApiABC

logger = logging.getLogger(__name__)


class DiagramImportApi(DiagramImportApiABC):
    def __init__(self, mainController: StatusController,
                 dispImportController: DispImportController,
                 lookupImportController: LookupImportController,
                 branchUpdateController: BranchUpdateController):
        self._mainController = mainController
        self._dispImportController = dispImportController
        self._lookupImportController = lookupImportController
        self._branchUpdateController = branchUpdateController

    def shutdown(self):
        pass

    def importDisps(self, modelSetKey: str,
                    coordSetKey: str,
                    importGroupHash: str,
                    dispsEncodedPayload: bytes) -> Deferred:
        return self._dispImportController.importDisps(
            modelSetKey, coordSetKey,
            importGroupHash, dispsEncodedPayload
        )

    def importBranches(self, branchesEncodedPayload: bytes) -> Deferred:
        return self._branchUpdateController.importBranches(branchesEncodedPayload)

    def importLookups(self, modelSetKey: str, coordSetKey: str,
                      lookupTupleType: str, lookupTuples: List,
                      deleteOthers: bool = True,
                      updateExisting: bool = True) -> Deferred:
        return self._lookupImportController.importLookups(
            modelSetKey, coordSetKey, lookupTupleType, lookupTuples,
            deleteOthers, updateExisting
        )

    def getLookups(self, modelSetKey: str, coordSetKey: str,
                   lookupTupleType: str) -> Deferred:
        return self._lookupImportController.getLookups(
            modelSetKey, coordSetKey, lookupTupleType
        )



from typing import Dict

from vortex.Tuple import Tuple, addTupleType, TupleField

from peek_plugin_diagram._private.PluginNames import diagramTuplePrefix


@addTupleType
class BranchKeyToIdMapTuple(Tuple):
    """ Branch Key to ID Map Tuple

    This tuple is used by the UI to get the IDs for branches for the
    model compiler to enable branches.

    """
    __tupleType__ = diagramTuplePrefix + 'BranchKeyToIdMapTuple'

    # This field is server side only
    coordSetId: int = TupleField()
    keyIdMap: Dict[str, int] = TupleField()



"""coordSet.positionOnZoom

Peek Plugin Database Migration Script

Revision ID: a73574690197
Revises: 3b41dec74e46
Create Date: 2019-08-12 13:06:03.511202

"""

# revision identifiers, used by Alembic.
revision = 'a73574690197'
down_revision = '3b41dec74e46'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.add_column('ModelCoordSet', sa.Column('positionOnZoom', sa.Float(), server_default='2.0', nullable=True), schema='pl_diagram')

    op.execute('''
        UPDATE pl_diagram."ModelCoordSet"
        SET "positionOnZoom"=2.0
    ''')

    op.alter_column('ModelCoordSet', 'positionOnZoom', type_=sa.Float(), nullable=False, schema='pl_diagram')


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('ModelCoordSet', 'positionOnZoom', schema='pl_diagram')
    # ### end Alembic commands ###


import os
import shutil
from setuptools import setup

from setuptools import find_packages

#
# Modify these values to fork a new plugin
#
author = "Synerty"
author_email = 'contact@synerty.com'
py_package_name = "peek_plugin_diagram"
pip_package_name = py_package_name.replace('_', '-')
package_version = '2.1.6'
description = 'Peek Plugin Diagram - My first enhancement.'

download_url = 'https://bitbucket.org/synerty/%s/get/%s.zip'
download_url %= pip_package_name, package_version
url = 'https://bitbucket.org/synerty/%s' % pip_package_name

#
# Everything below should be ok
#
egg_info = "%s.egg-info" % pip_package_name
if os.path.isdir(egg_info):
    shutil.rmtree(egg_info)

if os.path.isfile('MANIFEST'):
    os.remove('MANIFEST')

includePathContents = ('ns-assets',)
excludePathContains = ('__pycache__', 'node_modules', 'platforms', 'dist')
excludeFilesEndWith = ('.pyc', '.js', '.js.map', '.lastHash')
excludeFilesStartWith = ()


def find_package_files():
    paths = []
    for (path, directories, filenames) in os.walk(py_package_name):
        # Keep the dist-web dir no matter what
        if [e for e in includePathContents if e in path]:
            for filename in filenames:
                paths.append(os.path.join(path[len(py_package_name) + 1:], filename))
            continue

        if [e for e in excludePathContains if e in path]:
            continue

        for filename in filenames:
            if [e for e in excludeFilesEndWith if filename.endswith(e)]:
                continue

            if [e for e in excludeFilesStartWith if filename.startswith(e)]:
                continue

            paths.append(os.path.join(path[len(py_package_name) + 1:], filename))

    return paths


package_files = find_package_files()

setup(
    name=pip_package_name,
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    package_data={'': package_files},
    install_requires=['ujson', 'peek-plugin-base', 'Numpy'],
    version=package_version,
    description=description,
    author=author,
    author_email=author_email,
    url=url,
    download_url=download_url,
    keywords=['Peek', 'Python', 'Platform', 'synerty'],
    classifiers=[],
)


"""Fixed importHash indexes for lookups

Peek Plugin Database Migration Script

Revision ID: 9e7ca1acd6be
Revises: 7eed363f5df5
Create Date: 2017-08-12 12:30:43.048362

"""

# revision identifiers, used by Alembic.
revision = '9e7ca1acd6be'
down_revision = '7eed363f5df5'
branch_labels = None
depends_on = None

from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('idx_DispColor_importHash', table_name='DispColor', schema='pl_diagram')
    op.drop_index('idx_DispColor_modelSetId', table_name='DispColor', schema='pl_diagram')

    op.drop_index('idx_DispLayer_modelSetId', table_name='DispLayer', schema='pl_diagram')

    op.drop_index('idx_DispLevel_coordSetId', table_name='DispLevel', schema='pl_diagram')

    op.drop_index('idx_DispLineStyle_importHash', table_name='DispLineStyle', schema='pl_diagram')
    op.drop_index('idx_DispLineStyle_modelSetId', table_name='DispLineStyle', schema='pl_diagram')

    op.drop_index('idx_DispTextStyle_importHash', table_name='DispTextStyle', schema='pl_diagram')
    op.drop_index('idx_DispTextStyle_modelSetId', table_name='DispTextStyle', schema='pl_diagram')

    op.create_index('idx_DispColor_importHash', 'DispColor', ['modelSetId', 'importHash'], unique=True, schema='pl_diagram')
    op.create_index('idx_DispColor_modelSetId', 'DispColor', ['modelSetId'], unique=False, schema='pl_diagram')

    op.create_index('idx_DispLayer_importHash', 'DispLayer', ['modelSetId', 'importHash'], unique=True, schema='pl_diagram')
    op.create_index('idx_DispLayer_modelSetId', 'DispLayer', ['modelSetId'], unique=False, schema='pl_diagram')

    op.create_index('idx_DispLevel_coordSetId', 'DispLevel', ['coordSetId'], unique=False, schema='pl_diagram')
    op.create_index('idx_DispLevel_importHash', 'DispLevel', ['coordSetId', 'importHash'], unique=True, schema='pl_diagram')

    op.create_index('idx_DispLineStyle_importHash', 'DispLineStyle', ['modelSetId', 'importHash'], unique=True, schema='pl_diagram')
    op.create_index('idx_DispLineStyle_modelSetId', 'DispLineStyle', ['modelSetId'], unique=False, schema='pl_diagram')

    op.create_index('idx_DispTextStyle_importHash', 'DispTextStyle', ['modelSetId', 'importHash'], unique=True, schema='pl_diagram')
    op.create_index('idx_DispTextStyle_modelSetId', 'DispTextStyle', ['modelSetId'], unique=False, schema='pl_diagram')
    # ### end Alembic commands ###


def downgrade():
    raise NotImplementedError("Downgrade is not implemented")



from typing import Type

from peek_plugin_base.agent.PluginAgentEntryHookABC import PluginAgentEntryHookABC
from peek_plugin_base.client.PluginClientEntryHookABC import PluginClientEntryHookABC
from peek_plugin_base.server.PluginServerEntryHookABC import PluginServerEntryHookABC
from peek_plugin_base.worker.PluginWorkerEntryHookABC import PluginWorkerEntryHookABC

__version__ = '2.1.6'


def peekServerEntryHook() -> Type[PluginServerEntryHookABC]:
    from ._private.server.ServerEntryHook import ServerEntryHook
    return ServerEntryHook


def peekAgentEntryHook() -> Type[PluginAgentEntryHookABC]:
    from ._private.agent.AgentEntryHook import AgentEntryHook
    return AgentEntryHook


def peekWorkerEntryHook() -> Type[PluginWorkerEntryHookABC]:
    from ._private.worker.WorkerEntryHook import WorkerEntryHook
    return WorkerEntryHook


def peekClientEntryHook() -> Type[PluginClientEntryHookABC]:
    from ._private.client.ClientEntryHook import ClientEntryHook
    return ClientEntryHook



import logging
from typing import Dict, List

from twisted.internet.defer import inlineCallbacks

from peek_plugin_diagram._private.PluginNames import diagramFilt
from peek_plugin_diagram._private.server.client_handlers.ClientLocationIndexLoaderRpc import \
    ClientLocationIndexLoaderRpc
from peek_plugin_diagram._private.tuples.location_index.EncodedLocationIndexTuple import \
    EncodedLocationIndexTuple
from peek_plugin_diagram._private.tuples.location_index.LocationIndexTuple import \
    LocationIndexTuple
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope

logger = logging.getLogger(__name__)

clientLocationIndexUpdateFromServerFilt = dict(key="clientLocationIndexUpdateFromServer")
clientLocationIndexUpdateFromServerFilt.update(diagramFilt)


class LocationIndexCacheController:
    """ Disp Key Cache Controller

    The locationIndex cache controller stores all the locationIndexs in memory, allowing fast access from
    the mobile and desktop devices.

    """

    LOAD_CHUNK = 50

    def __init__(self, clientId: str):
        self._clientId = clientId
        self._cacheHandler = None
        self._cache: Dict[str, EncodedLocationIndexTuple] = {}
        self._locationIndexKeys = set()

        self._endpoint = PayloadEndpoint(clientLocationIndexUpdateFromServerFilt,
                                         self._processLocationIndexPayload)

    def setLocationIndexCacheHandler(self, locationIndexCacheHandler):
        self._cacheHandler = locationIndexCacheHandler

    @inlineCallbacks
    def start(self):
        yield self.reloadCache()

    def shutdown(self):
        self._tupleObservable = None

        self._endpoint.shutdown()
        self._endpoint = None

        self._cache = {}

    @inlineCallbacks
    def reloadCache(self):
        self._cache = {}
        self._locationIndexKeys = set()

        offset = 0
        while True:
            logger.info("Loading LocationIndex %s to %s" % (
                offset, offset + self.LOAD_CHUNK))
            locationIndexTuples = yield ClientLocationIndexLoaderRpc.loadLocationIndexes(
                offset, self.LOAD_CHUNK)
            if not locationIndexTuples:
                break

            self._loadLocationIndexIntoCache(locationIndexTuples)

            offset += self.LOAD_CHUNK

    @inlineCallbacks
    def _processLocationIndexPayload(self, payloadEnvelope: PayloadEnvelope, **kwargs):
        paylod = yield payloadEnvelope.decodePayloadDefer()
        locationIndexTuples: List[LocationIndexTuple] = paylod.tuples
        self._loadLocationIndexIntoCache(locationIndexTuples)

    def _loadLocationIndexIntoCache(self, locationIndexTuples: List[LocationIndexTuple]):
        indexBucketsUpdated: List[str] = []

        for t in locationIndexTuples:
            self._locationIndexKeys.add(t.indexBucket)

            if (not t.indexBucket in self._cache or
                    self._cache[t.indexBucket].lastUpdate != t.lastUpdate):
                self._cache[t.indexBucket] = t
                indexBucketsUpdated.append(t.indexBucket)

        logger.debug("Received locationIndex updates from server, %s",
                     indexBucketsUpdated)

        self._cacheHandler.notifyOfLocationIndexUpdate(indexBucketsUpdated)

    def locationIndex(self, indexBucket) -> EncodedLocationIndexTuple:
        return self._cache.get(indexBucket)

    def locationIndexKeys(self) -> List[str]:
        return list(self._locationIndexKeys)



"""added branch tables

Peek Plugin Database Migration Script

Revision ID: f76791d27f6d
Revises: 5e2f84d757f2
Create Date: 2019-02-04 19:34:07.351930

"""

# revision identifiers, used by Alembic.
revision = 'f76791d27f6d'
down_revision = '5e2f84d757f2'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('BranchIndexCompilerQueue',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('modelSetId', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('chunkKey', sa.String(), nullable=False),
    sa.ForeignKeyConstraint(['modelSetId'], ['pl_diagram.ModelSet.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id', 'chunkKey'),
    schema='pl_diagram'
    )
    op.create_table('BranchIndexEncodedChunk',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('modelSetId', sa.Integer(), nullable=False),
    sa.Column('chunkKey', sa.String(), nullable=False),
    sa.Column('encodedData', sa.LargeBinary(), nullable=False),
    sa.Column('encodedHash', sa.String(), nullable=False),
    sa.Column('lastUpdate', sa.String(), nullable=False),
    sa.ForeignKeyConstraint(['modelSetId'], ['pl_diagram.ModelSet.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id'),
    schema='pl_diagram'
    )
    op.create_index('idx_BranchIndexEnc_modelSetId_chunkKey', 'BranchIndexEncodedChunk', ['modelSetId', 'chunkKey'], unique=False, schema='pl_diagram')
    op.create_table('BranchIndex',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('coordSetId', sa.Integer(), nullable=False),
    sa.Column('importGroupHash', sa.String(), nullable=True),
    sa.Column('importHash', sa.String(), nullable=True),
    sa.Column('key', sa.String(), nullable=False),
    sa.Column('chunkKey', sa.String(), nullable=False),
    sa.Column('packedJson', sa.String(), nullable=False),
    sa.ForeignKeyConstraint(['coordSetId'], ['pl_diagram.ModelCoordSet.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id'),
    schema='pl_diagram'
    )
    op.create_index('idx_BranchIndex_chunkKey', 'BranchIndex', ['chunkKey'], unique=False, schema='pl_diagram')
    op.create_index('idx_BranchIndex_importGroupHash', 'BranchIndex', ['coordSetId', 'importGroupHash'], unique=False, schema='pl_diagram')
    op.create_index('idx_BranchIndex_importHash', 'BranchIndex', ['coordSetId', 'importHash'], unique=True, schema='pl_diagram')
    op.create_index('idx_BranchIndex_key', 'BranchIndex', ['coordSetId', 'key'], unique=True, schema='pl_diagram')
    op.create_table('BranchGridIndex',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('branchIndexId', sa.Integer(), nullable=False),
    sa.Column('coordSetId', sa.Integer(), nullable=False),
    sa.Column('gridKey', sa.String(), nullable=False),
    sa.ForeignKeyConstraint(['branchIndexId'], ['pl_diagram.BranchIndex.id'], ondelete='CASCADE'),
    sa.ForeignKeyConstraint(['coordSetId'], ['pl_diagram.ModelCoordSet.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id', 'coordSetId'),
    schema='pl_diagram'
    )
    op.create_index('idx_BranchGridIndex_coord_key', 'BranchGridIndex', ['coordSetId', 'branchIndexId', 'gridKey'], unique=True, schema='pl_diagram')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('idx_BranchGridIndex_coord_key', table_name='BranchGridIndex', schema='pl_diagram')
    op.drop_table('BranchGridIndex', schema='pl_diagram')
    op.drop_index('idx_BranchIndex_key', table_name='BranchIndex', schema='pl_diagram')
    op.drop_index('idx_BranchIndex_importGroupHash', table_name='BranchIndex', schema='pl_diagram')
    op.drop_index('idx_BranchIndex_chunkKey', table_name='BranchIndex', schema='pl_diagram')
    op.drop_table('BranchIndex', schema='pl_diagram')
    op.drop_index('idx_BranchIndexEnc_modelSetId_chunkKey', table_name='BranchIndexEncodedChunk', schema='pl_diagram')
    op.drop_table('BranchIndexEncodedChunk', schema='pl_diagram')
    op.drop_table('BranchIndexCompilerQueue', schema='pl_diagram')
    # ### end Alembic commands ###